"""
The `quizcomp.cli.gradescope` package contains tools for interacting with Gradescope.
"""

import sys

import quizcomp.util.cli

def main():
    return quizcomp.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())
