"""
The `quizcomp.cli.parse` package contains tools for parsing various objects/files from disk.
"""

import sys

import quizcomp.util.cli

def main():
    return quizcomp.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())
